""":mod:`dict2rel` is a Python package aimed at aiding situations where you
have highly nested JSON objects and want to either flatten them into a
single table or transform them into multiple tables where nested lists get
broken out into their own sheets; turning a dict into relational tables; and
providing the name.

The library is table-type agnostic and therefore ``pandas.DataFrame``,
``polars.DataFrame``, or any other provider can be used when constructing
the tables. The two previously mentioned types are automatically handled
when converting back from tables into JSON.

>>> from dict2rel import dict2rel, UnravelOptions
>>> tables = dict2rel(
...     {
...         "project": "Alpha-Prime",
...         "version": "1.0.3",
...         "config": {
...             "modules": [
...                 {
...                     "id": "A1",
...                     "status": "active",
...                     "settings": {
...                         "security": {
...                             "encryption_level": 5,
...                             "algorithms": [
...                                 "AES-256",
...                                 "SHA-512"
...                             ]
...                         }
...                     }
...                 },
...                 {
...                     "id": "B2",
...                     "status": "passive",
...                     "settings": {
...                         "security": {
...                             "encryption_level": 0,
...                             "algorithms": [
...                                 "AES-256"
...                             ]
...                         }
...                     }
...                 }
...             ]
...         }
...     },
...     UnravelOptions(marker="Expanded {len} results to {sheet}")
... )
>>> tables
{
    "*": pd.DataFrame([...]),
    "*.config.modules": pd.DataFrame([...]),
    "*.config.modules.*.settings.security.algorithms": pd.DataFrame([...])
}

Where the tables in the example above are as follows.

``*``:
    =========== ======= ======================================== ===
    project     version config.modules                           _id
    =========== ======= ======================================== ===
    Alpha-Prime 1.0.3   Expanded 2 results to \\*.config.modules  0
    =========== ======= ======================================== ===

``*.config.modules``:
    == ======= ================================== ===================================================================== ===================
    id status  settings.security.encryption_level settings.security.algorithms                                          _id
    == ======= ================================== ===================================================================== ===================
    A1 active  5                                  Expanded 2 results to \\*.config.modules.settings.security.algorithms  0.config.modules.0
    B2 passive 0                                  Expanded 1 results to \\*.config.modules.settings.security.algorithms  0.config.modules.1
    == ======= ================================== ===================================================================== ===================

``*.config.modules.*.settings.security.algorithm``:
    =======  =================================================
    _value   _id
    =======  =================================================
    AES-256  0.config.modules.0.settings.security.algorithms.0
    SHA-512  0.config.modules.0.settings.security.algorithms.1
    AES-256  0.config.modules.1.settings.security.algorithms.0
    =======  =================================================

These tables can be converted back to the original JSON by applying :func:`rel2dict`.

>>> from dict2rel import rel2dict
>>> rel2dict(tables)
{
    # original value
}

:mod:`dict2rel` also provides functions for converting JSON into a single
table and then back to JSON with :func:`~dict2rel.flatten` and
:func:`~dict2rel.inflate`.
"""

# SPDX-FileCopyrightText: 2025-present Jacob Morris <blendingjake@gmail.com>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Iterable, TypeVar

if TYPE_CHECKING:
    from dict2rel._types import JsonObject, Row

from dict2rel.__about__ import __version__
from dict2rel._errors import ToRowsRequiredError
from dict2rel._ravel import inflate as _inflate
from dict2rel._ravel import ravel as _ravel
from dict2rel._types import UnravelOptions
from dict2rel._unravel import flattener as _flattener
from dict2rel._unravel import unravel as _unravel

__all__ = [
    "__version__",
    "dict2rel",
    "flatten",
    "inflate",
    "rel2dict",
    "ToRowsRequiredError",
    "UnravelOptions",
]
P = TypeVar("P")


def dict2rel(
    obj: list[JsonObject] | JsonObject,
    provider: Callable[[list[Row]], P],
    options: UnravelOptions | None = None,
) -> dict[str, P]:
    """Take a list of (or single) JSON object(s) and convert them to tables using
    the provider of your choice to construct the tables (like Polars, Pandas, etc.).
    Nested arrays of JSON objects will be broken out into their own tables while
    nested objects will be flattened inline. ``options`` can be provided to do things
    like place a marker whenever a list is expanded to a new table instead of dropping
    the column.

    :func:`rel2dict` can be used to convert the results of this function back to
    ``obj``, such that ``rel2dict(dict2rel(obj, ...)) == obj``.

    >>> dict2rel([
    ...     {
    ...         "name": {
    ...             "first": "John",
    ...             "last": "Smith"
    ...         },
    ...         "phones": [
    ...             {
    ...                 "country": "USA",
    ...                 "number": "1234567890"
    ...             },
    ...             {
    ...                 "country": "ESP",
    ...                 "number": "987654321"
    ...             }
    ...         ]
    ...     }
    ... ], pd.DataFrame)
    {
        "*": pd.DataFrame([
            {
                "_id": "0",
                "name.first": "John",
                "name.last": "Smith"
            }
        ]),
        "*.phones": pd.DataFrame([
            {
                "_id": "0.phones.0",
                "country": "USA",
                "number": "1234567890"
            },
            {
                "_id": "0.phones.1",
                "country": "ESP",
                "number": "987654321"
            }
        ])
    }

    :param obj: A :attr:`JSONObject` or list of them
    :param provider: A function which converts a list of rows into a table. Typically,
        this will be a value like ``pandas.DataFrame`` or ``polars.DataFrame``, but
        can be an identity lambda which will return the results as lists of dictionaries.
    :param options: Options to configure how ``obj`` is unraveled, like whether to place
        markers whenever a column is a list which gets expanded to its own table.
    """
    objs = [obj] if isinstance(obj, dict) else obj
    rows: dict[str, list[Row]] = {}
    for sheet, data in _unravel(objs, options or UnravelOptions()):
        if sheet not in rows:
            rows[sheet] = []

        rows[sheet].append(data)

    return {sheet: provider(rs) for sheet, rs in rows.items()}


def flatten(
    obj: list[JsonObject] | JsonObject, provider: Callable[[list[Row]], P]
) -> P:
    """Take a list of objects, or a single dict, and flatten it into a single sheet.
    Unlike :func:`dict2rel`, nested lists are kept on the primary sheet and
    provided unique column names.

    :func:`inflate` can be used to reverse this process such that
    ``inflate(flatten(obj, ...)) == obj``.

    >>> from dict2rel import flatten
    >>> flatten([
    ...     {
    ...         "name": {
    ...             "first": "John",
    ...             "last": "Smith"
    ...         },
    ...         "phones": [
    ...             {
    ...                 "country": "USA",
    ...                 "number": "1234567890"
    ...             },
    ...             {
    ...                 "country": "ESP",
    ...                 "number": "987654321"
    ...             }
    ...         ]
    ...     }
    ... ], pl.DataFrame)
    pl.DataFrame([
        {
            "_id": "0",
            "name.first": "John",
            "name.last": "Smith",
            "phones.0.country": "USA",
            "phones.0.number": "1234567890",
            "phones.1.country": "ESP",
            "phones.1.number": "987654321"
        }
    ])

    :param obj: A :attr:`JSONObject` or list of them
    :param provider: A function which converts a list of rows into a table. Typically,
        this will be a value like ``pandas.DataFrame`` or ``polars.DataFrame``, but
        can be an identity lambda which will return the results as list of dictionaries.
    """
    objs = [obj] if isinstance(obj, dict) else obj
    return provider(list(_flattener(objs)))


def inflate(
    table: P, to_rows: Callable[[P], Iterable[Row]] | None = None
) -> list[JsonObject]:
    """Undo :func:`flatten` and take a sheet with nesting represented by column
    names and inflate it back to a list of dictionaries with actual nesting.

    >>> from dict2rel import inflate
    >>> inflate(pl.DataFrame([
    ...     {
    ...         "name": "Bravo",
    ...         "version.major": "1",
    ...         "version.minor": "0",
    ...         "version.patch": "12",
    ...         "releases.0.date": "2025-02-12",
    ...         "releases.0.version": "0.0.1",
    ...         "releases.1.date": "2025-02-18",
    ...         "releases.1.version": "0.1.0"
    ...     }
    ... ]))
    [{
        'name': 'Bravo',
        'version': {
            'major': '1',
            'minor': '0',
            'patch': '12'
        },
        'releases': [
            {'date': '2025-02-12', 'version': '0.0.1'},
            {'date': '2025-02-18', 'version': '0.1.0'}
        ]
    }]

    :param table: The table to inflate. This can either be a list of dictionaries,
        or a table such as ``pandas.DataFrame`` or ``polars.DataFrame``.
    :param to_rows: A function to convert the table data to dictionaries. This is only
        needed if the data isn't already in that format or the tables are a datatype
        other than ``pandas.DataFrame`` or ``polars.DataFrame``.
    :raise ToRowsRequiredError: If any of the tables aren't lists or a known
        table-type like ``pandas.DataFrame`` or ``polars.DataFrame``.
    """
    rows: Iterable[Row] = _to_rows(to_rows)(table)
    return _inflate(rows)


def rel2dict(
    tables: dict[str, P],
    to_rows: Callable[[P], Iterable[Row]] | None = None,
) -> list[JsonObject]:
    """Take a mapping of tables, likely produced by :func:`dict2rel`, and
    reconstruct the nested JSON from them. The tables themselves can be objects
    like ``pandas.DataFrame`` or ``polars.DataFrame``, or other table-types if
    ``to_rows`` is provided.

    >>> from dict2rel import rel2dict
    >>> rel2dict(
    ...     {
    ...         "*": pl.DataFrame([
    ...             {
    ...                 "_id": "0",
    ...                 "name": "Acme Corp.",
    ...                 "state": "AZ",
    ...                 "board": "4 board members in *.board"
    ...             },
    ...             {
    ...                 "_id": "1",
    ...                 "name": "ZZZ Consulting",
    ...                 "state": "NY",
    ...                 "board": "2 board members in *.board"
    ...             }
    ...         ]),
    ...         "*.board": pl.DataFrame([
    ...             {
    ...                 "_id": "0.board.0",
    ...                 "name": "Wile E. Coyote"
    ...             },
    ...             {
    ...                 "_id": "0.board.1",
    ...                 "name": "Someone Else"
    ...             },
    ...             {
    ...                 "_id": "1.board.0",
    ...                 "name": "Leonhard Euler"
    ...             },
    ...             {
    ...                 "_id": "1.board.1",
    ...                 "name": "Carl Gauss"
    ...             }
    ...         ])
    ...     }
    ... )
    [
        {
            'name': 'Acme Corp.',
            'state': 'AZ',
            'board': [
                {'name': 'Wile E. Coyote'},
                {'name': 'Someone Else'}
            ]
        },
        {
            'name': 'ZZZ Consulting',
            'state': 'NY',
            'board': [
                {'name': 'Leonhard Euler'},
                {'name': 'Carl Gauss'}
            ]
        }
    ]

    :param tables: A mapping of table names to table data
    :param to_rows: A function to convert the table data to dictionaries. This is only
        needed if the data isn't already in that format or the tables are a datatype
        other than ``pandas.DataFrame`` or ``polars.DataFrame``.
    :raise ToRowsRequiredError: If any of the tables aren't lists or a known
        table-type like ``pandas.DataFrame`` or ``polars.DataFrame``.
    """
    true_to_rows = _to_rows(to_rows)
    return _ravel({sheet: true_to_rows(rows) for sheet, rows in tables.items()})


def _to_rows(
    to_rows: Callable[[P], Iterable[Row]] | None,
) -> Callable[[P], Iterable[Row]]:
    """Take an optional to_rows converter and produce a new converter
    which will fallback and fill handling for pandas and polars.
    """

    def worker(sheet: P) -> Iterable[Row]:
        if to_rows:
            return to_rows(sheet)

        tp = type(sheet)
        if tp.__name__ == "DataFrame":
            if "pandas" in tp.__module__:
                return (row for _, row in sheet.iterrows())
            if "polars" in tp.__module__:
                return sheet.rows(named=True)
        if isinstance(sheet, list):
            return sheet

        raise ToRowsRequiredError

    return worker



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Union

FieldPath = List[Union[int, str]]

JsonPrimitive = Union[int, float, str, bool, None]
JsonValue = Union[JsonPrimitive, List["JsonValue"], "JsonObject"]
JsonObject = Dict[str, JsonValue]
Row = Dict[str, JsonPrimitive]

ID_SENTINEL = "_id"
VALUE_SENTINEL = "_value"


@dataclass
class UnravelOptions:
    """Options for configuring how an object is unraveled and
    expanded into one or more tables. Used by :func:`dict2rel.dict2rel`.
    """

    marker: str | None = None
    """The value, if any, which will be placed in a column when
    the value was a list and therefore got expanded to its own table.
    By default, the column is not included.

    String interpolation is supported and the provided values are:

    * ``field: str`` - the name of the field being expanded
    * ``id: str`` - the ``_id`` of the current row
    * ``len: int`` - the length of the nested list
    * ``sheet: str`` - the name of the sheet where the nested values
      were placed

    An example marker value would be: ``"{len} values placed in {sheet}"``.
    """



import polars as pl

from dict2rel import inflate


def test_inflate_with_polars():
    """Test inflation example using a Polars DataFrame as the source"""
    tbl = pl.DataFrame(
        [
            {
                "name": "Bravo",
                "version.major": "1",
                "version.minor": "0",
                "version.patch": "12",
                "releases.0.date": "2025-02-12",
                "releases.0.version": "0.0.1",
                "releases.1.date": "2025-02-18",
                "releases.1.version": "0.1.0",
            }
        ]
    )

    results = inflate(tbl)

    assert len(results) == 1
    assert "releases" in results[0]
    assert len(results[0]["releases"]) == 2
    assert "version" in results[0]
    assert results[0]["version"]["major"] == "1"



DOCSTRING_EXAMPLE_1 = {
    "project": "Alpha-Prime",
    "version": "1.0.3",
    "config": {
        "modules": [
            {
                "id": "A1",
                "status": "active",
                "settings": {
                    "security": {
                        "encryption_level": 5,
                        "algorithms": ["AES-256", "SHA-512"],
                    }
                },
            },
            {
                "id": "B2",
                "status": "passive",
                "settings": {
                    "security": {"encryption_level": 0, "algorithms": ["AES-256"]}
                },
            },
        ]
    },
}

EXAMPLE_SMALL = {
    "name": {"first": "John", "last": "Smith", "suffixes": ["MD", "JD"]},
    "phones": [
        {
            "country": "USA",
            "number": "1234567890",
        },
        {"country": "ESP", "number": "987654321"},
    ],
    "matrix": [[1, 2], [3, 4]],
}



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent.parent.parent / "src"))

from dict2rel import __version__

project = "dict2rel"
copyright = "2025, Jacob Morris"
author = "Jacob Morris"
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc", "sphinx_toolbox.more_autodoc.no_docstring"]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]



from __future__ import annotations

from typing import Iterator, NamedTuple

from dict2rel._types import (
    ID_SENTINEL,
    VALUE_SENTINEL,
    FieldPath,
    JsonObject,
    JsonValue,
    Row,
    UnravelOptions,
)

PartialMarker = NamedTuple(
    "PartialMarker", (("field", str), ("len", int), ("path", FieldPath))
)


def _determine_sheet_name(parts: FieldPath) -> str:
    sheet_parts = ["*" if isinstance(part, int) else part for part in parts]

    if len(sheet_parts) > 1 and sheet_parts[-1] == "*":
        sheet_parts = sheet_parts[:-1]

    return ".".join(sheet_parts)


def flattener(obj: list[JsonObject]) -> Iterator[Row]:
    """Take a list of :attr:`JsonObject` and flatten all nesting to a
    dictionary where the values will be singletons. Yield out those
    flattened rows.
    """
    root_objs: dict[int | str, list[tuple[FieldPath, Row]]] = {}
    for fp, true_obj in _unravel(obj, [], UnravelOptions()):
        if fp[0] not in root_objs:
            root_objs[fp[0]] = []

        root_objs[fp[0]].append((fp, true_obj))

    for root_id, objs in root_objs.items():
        sorted_objs = sorted(objs, key=lambda x: x[0])
        root_obj: Row = {ID_SENTINEL: str(root_id), **sorted_objs[0][1]}

        for child_path, child_obj in sorted_objs[1:]:
            # Remove first part of path as it will be root_id
            child_id = ".".join(map(str, child_path[1:]))

            if len(child_obj) == 1 and VALUE_SENTINEL in child_obj:
                root_obj[child_id] = child_obj[VALUE_SENTINEL]
            else:
                root_obj.update({f"{child_id}.{k}": v for k, v in child_obj.items()})

        yield root_obj


def unravel(
    obj: list[JsonObject], options: UnravelOptions
) -> Iterator[tuple[str, Row]]:
    for fp, true_obj in _unravel(obj, [], options):
        _id = ".".join(map(str, fp))
        true_obj[ID_SENTINEL] = _id
        sheet = _determine_sheet_name(fp)

        for k, v in tuple(true_obj.items()):
            if isinstance(v, PartialMarker) and options.marker:
                true_obj[k] = options.marker.format(
                    field=v.field,
                    id=_id,
                    len=v.len,
                    sheet=_determine_sheet_name(v.path),
                )

        yield sheet, true_obj


def _unravel(
    obj: JsonValue, path: FieldPath, options: UnravelOptions
) -> Iterator[tuple[FieldPath, Row]]:
    if isinstance(obj, list):
        for i, v in enumerate(obj):
            yield from _unravel(v, [*path, i], options)
    elif isinstance(obj, dict):
        new_obj: Row = {}
        for k, v in obj.items():
            if isinstance(v, list):
                new_path: FieldPath = [*path, k]
                yield from _unravel(v, new_path, options)

                if options.marker is not None:
                    new_obj[k] = PartialMarker(field=k, len=len(v), path=new_path)
            elif isinstance(v, dict):
                for fp, nested_obj in _unravel(v, [*path, k], options):
                    if all(isinstance(part, str) for part in fp[len(path) + 1 :]):
                        for kk, vv in nested_obj.items():
                            new_obj[
                                ".".join([k, *list(map(str, fp[len(path) + 1 :])), kk])
                            ] = vv
                    else:
                        yield fp, nested_obj
            else:
                new_obj[k] = v

        yield path, new_obj
    else:
        yield path, {VALUE_SENTINEL: obj}



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from dict2rel._types import JsonObject, JsonValue, Row

from dict2rel._types import ID_SENTINEL, VALUE_SENTINEL


def _parse_int_and_pad_parent(val: str, parent: list[JsonValue]) -> int:
    """Parse a string which is an int and then pad the length
    of parent to ensure the int will be a valid index.
    """
    value = int(val)
    parent.extend(None for _ in range(max(0, value - len(parent) + 1)))
    return value


def inflate(rows: Iterable[Row]) -> list[JsonObject]:
    inflated_rows: list[JsonObject] = []
    for row in rows:
        new_obj: JsonObject = {}
        for k, v in row.items():
            if k == ID_SENTINEL:
                continue

            if "." in k:
                _rebuild_nesting_and_place_value(new_obj, tuple(k.split(".")), v)
            else:
                new_obj[k] = v

        inflated_rows.append(new_obj)
    return inflated_rows


def ravel(tables: dict[str, Iterable[Row]]) -> list[JsonObject]:
    # First, we need to reconstruct any nested dicts on all rows.
    id_to_value: dict[tuple[str, ...], JsonValue] = {}
    for rows in tables.values():
        for row in rows:
            id_to_value[tuple(row[ID_SENTINEL].split("."))] = _rebuild_dicts(row)

    # Figure out what is the most nested id and work from there up,
    # reconstructing the original values.
    most_nested = sorted(id_to_value, reverse=True)
    for parts in most_nested:
        if len(parts) <= 1:
            continue

        # Find the longest prefix id which exists in the lookup
        longest = (parts[0],)  # this has to be there
        for i in range(len(parts) - 1):
            if parts[:i] in id_to_value:
                longest = parts[:i]

        _rebuild_nesting_and_place_value(
            id_to_value[longest], parts[len(longest) :], id_to_value[parts]
        )

        # We've added this thing to its parent, so we can remove it
        # from the index. The bonus of this is that anything left at the
        # end must be a root/original row.
        del id_to_value[parts]

    return list(id_to_value.values())


def _rebuild_dicts(row: Row) -> JsonValue:
    """Take a row and rebuild any nested dicts that are present, remove
    the _id field, and handle any value-only rows. Rebuilding rows involves
    taking "name.first" and changing it back into {"name": {"first": ...}}.
    """
    # _id has to be there, so this must be a "value" row
    if len(row) == 2 and VALUE_SENTINEL in row:
        return row[VALUE_SENTINEL]

    new_obj: JsonObject = {}
    for key, value in row.items():
        if "." in key:
            parts = key.split(".")
            pointer = new_obj

            # Build any missing levels
            for part in parts[:-1]:
                if part not in pointer:
                    pointer[part] = {}

                pointer = pointer[part]

            # Assign the value to that last part
            pointer[parts[-1]] = value
        elif key != ID_SENTINEL:
            new_obj[key] = value

    return new_obj


def _rebuild_nesting_and_place_value(
    parent: JsonValue, path_to_build: tuple[str, ...], value_at_path: JsonValue
) -> JsonValue:
    """Build any missing layers referenced by ``path_to_build`` in
    ``parent``. Once rebuilt, place ``value_at_path`` there.

    >>> _rebuild_nesting_and_place_value(
    ...     {
    ...         "foo": "bar"
    ...     },
    ...     ["path1", "0", "path2"],
    ...     "the value"
    ... )
    {
        "foo": "bar",
        "path1": [
            {
                "path2": "the value"
            }
        ]
    }
    """
    # Build the layers that are missing between the closest parent
    # we found in the index and what is specified by this _id.
    for i in range(len(path_to_build) - 1):
        # The type of the current part being added is based on the
        # next part. If the next part is a numeric value, then this
        # thing we're adding must be a list. If it is a list, then make
        # sure it is long enough to handle the index we're setting.
        cur = path_to_build[i]
        _next = path_to_build[i + 1]
        next_type = list if _next.isdigit() else dict

        if cur.isdigit():
            value = _parse_int_and_pad_parent(cur, parent)
            if not isinstance(parent[value], next_type):
                parent[value] = next_type()

            parent = parent[value]
        else:
            if cur not in parent or not isinstance(parent[cur], next_type):
                parent[cur] = next_type()

            parent = parent[cur]

    # We've built the intervening levels, so we can go ahead and
    # assign the value of interest to what we've built.
    if path_to_build[-1].isdigit():
        value = _parse_int_and_pad_parent(path_to_build[-1], parent)
        parent[value] = value_at_path
    else:
        parent[path_to_build[-1]] = value_at_path



import pandas as pd
import polars as pl
import pytest

from dict2rel import ToRowsRequiredError, UnravelOptions, dict2rel, rel2dict

from .data import DOCSTRING_EXAMPLE_1, EXAMPLE_SMALL


def test_basic():
    """Test basic JSON -> tables"""
    tables = dict2rel([EXAMPLE_SMALL], lambda x: x)
    assert len(tables) == 4

    assert "*" in tables
    assert len(tables["*"]) == 1
    assert tables["*"][0]["name.first"] == EXAMPLE_SMALL["name"]["first"]

    assert "*.name.suffixes" in tables
    assert "_value" in tables["*.name.suffixes"][0]


@pytest.mark.parametrize("provider", [pd.DataFrame, pl.DataFrame])
def test_basic_with_providers(provider):
    """Test basic JSON -> tables with various providers"""
    tables = dict2rel(EXAMPLE_SMALL, provider)

    assert all(isinstance(v, provider) for v in tables.values())

    primary = tables["*"]
    assert "name.first" in primary
    assert primary["name.first"].to_list() == [EXAMPLE_SMALL["name"]["first"]]


def test_docstring_example_1():
    """Verify that the docstring example produces the expected result."""
    tables = dict2rel(
        DOCSTRING_EXAMPLE_1,
        pd.DataFrame,
        UnravelOptions(marker="Expanded {len} results to {sheet}"),
    )

    assert len(tables) == 3

    tables_and_size = [
        ("*", 1),
        ("*.config.modules", 2),
        ("*.config.modules.*.settings.security.algorithms", 3),
    ]
    for t, s in tables_and_size:
        assert t in tables
        assert len(tables[t]) == s


def test_rel2dict_basic():
    """A basic example ensuring rel2dict works with polars and data
    which didn't originate with dict2rel.
    """
    objs = rel2dict(
        {
            "*": pl.DataFrame(
                [
                    {
                        "_id": "0",
                        "name": "Acme Corp.",
                        "state": "AZ",
                        "board": "4 board members in *.board",
                    },
                    {
                        "_id": "1",
                        "name": "ZZZ Consulting",
                        "state": "NY",
                        "board": "2 board members in *.board",
                    },
                ]
            ),
            "*.board": pl.DataFrame(
                [
                    {"_id": "0.board.0", "name": "Wile E. Coyote"},
                    {"_id": "0.board.1", "name": "Someone Else"},
                    {"_id": "1.board.0", "name": "Leonhard Euler"},
                    {"_id": "1.board.1", "name": "Carl Gauss"},
                ]
            ),
        }
    )

    assert len(objs) == 2

    assert "board" in objs[0]
    assert len(objs[0]["board"]) == 2
    assert objs[0]["board"][1]["name"] == "Someone Else"

    assert "board" in objs[1]
    assert len(objs[1]["board"]) == 2
    assert objs[1]["board"][0]["name"] == "Leonhard Euler"


def test_reraveling_data_with_markers():
    """Verify that original data can be reconstructed even
    when markers were placed.
    """
    tables = dict2rel(
        EXAMPLE_SMALL, lambda x: x, UnravelOptions(marker="Expanded {len} values")
    )
    assert rel2dict(tables) == [EXAMPLE_SMALL]


@pytest.mark.parametrize(
    ("provider", "to_rows"),
    [
        (pd.DataFrame, None),
        (pl.DataFrame, None),
        (lambda x: x, None),
        (lambda x: x, lambda x: x),
    ],
)
def test_to_row_options(provider, to_rows):
    """Verify all of the ways to_rows can be provided"""
    tables = dict2rel(EXAMPLE_SMALL, provider)
    og = rel2dict(tables, to_rows)

    assert og[0] == EXAMPLE_SMALL


def test_to_row_not_provider():
    """Verify that an error is thrown when to_rows is not provided
    and the table-type is not known/supported.
    """

    class MyTable(pd.DataFrame):
        pass

    tables = dict2rel(EXAMPLE_SMALL, MyTable)
    with pytest.raises(ToRowsRequiredError):
        rel2dict(tables)


def test_to_tables_and_back_basic():
    """Test JSON -> tables -> JSON"""
    tables = dict2rel([EXAMPLE_SMALL], lambda x: x)
    og = rel2dict(tables)

    assert og == [EXAMPLE_SMALL]


@pytest.mark.parametrize(
    "provider",
    [pd.DataFrame, pl.DataFrame],
)
def test_to_tables_and_back_providers(provider):
    """Test JSON -> provider tables -> JSON"""
    tables = dict2rel(EXAMPLE_SMALL, provider)
    og = rel2dict(tables)
    assert og == [EXAMPLE_SMALL]


def test_with_unravel_options():
    """Test custom marker language"""
    fmt = "{field} had {len} values placed in {sheet}"
    tables = dict2rel(EXAMPLE_SMALL, lambda x: x, UnravelOptions(marker=fmt))

    assert "*" in tables
    assert "*.phones" in tables

    primary = tables["*"]
    assert len(primary) == 1
    assert "phones" in primary[0]
    assert primary[0]["phones"] == fmt.format(field="phones", len=2, sheet="*.phones")
